/*******************************************************************************
 *
 * Driver for SSD1963 Display written as CFunctions
 *
 * (c) Peter Mather 2015 with acknowledgements to Peter Carnegie & Geoff Graham
 * 
 *
 * This CFunction MUST be compiled with Optimization Level 1, -O1
 * -O2,-O3,-Os will compile successfully, but generate exceptions at runtime.
 *
 * When Generating the CFunction, use MERGE CFunction mode, and name the CFunction
 * SSD1963_V44
 *
 * Entry point is function long long main(long long *MyAddress,
 *                                        long long *RST,
 *                                        long long *orientation
 *                                        long long *size)
 *
 * V1.0     2015-07-21 Peter Mather
 * 1.01     Change orientation codes to 1:landscape, 2:portrait, 3:rlandscape, 4:rportrait
 * 1.02     Use fixed PortC pinout for maximum performance
 * 1.03     Stripped routines for use as loadable LIBRARY
 * 1.04     Conditional compilation for 44 and 64-pin parts
 * 1.1      Convert to Nathan approach
 * 1.11     Minor change for compatibility with touch and restricting page usage
 * 1.2      Implement RGB565, add-in triangle drawing read/write
 ******************************************************************************/
#define Version 111     //Version 1.11
#include <stdarg.h>
#include <xc.h>			   // Required for SFR defs
#include <sys/attribs.h>   // Required to use __longramfunc__
#include "../CFunctions.h"
#define MX470 //Comment out for compilation for 44-pin part
#define pin100
#define RGB565
#define aport *(volatile unsigned int *)(0xbf886030) //latch registers
#define aclrport *(volatile unsigned int *)(0xbf886034) //latch registers
#define asetport *(volatile unsigned int *)(0xbf886038) //latch registers
#define aread *(volatile unsigned int *)(0xbf886020) //latch registers
#define atrisinv *(volatile unsigned int *)(0xbf88601C) //latch registers
#define bclrport *(volatile unsigned int *)(0xbf886134) //latch registers
#define bsetport *(volatile unsigned int *)(0xbf886138) //latch registers
#define cport *(volatile unsigned int *)(0xbf886230) //latch registers
#define cread *(volatile unsigned int *)(0xbf886220) //latch registers
#define ctrisinv *(volatile unsigned int *)(0xbf88621C) //latch registers
#define cclrport *(volatile unsigned int *)(0xbf886234) //latch registers
#define csetport *(volatile unsigned int *)(0xbf886238) //latch registers
#define eport *(volatile unsigned int *)(0xbf886430) //latch registers
#define eread *(volatile unsigned int *)(0xbf886420) //latch registers
#define etrisinv *(volatile unsigned int *)(0xbf88641C) //latch registers
#define eclrport *(volatile unsigned int *)(0xbf886434) //latch registers
#define esetport *(volatile unsigned int *)(0xbf886438) //latch registers
#define DEVID (*(volatile unsigned int *)0xBF80F220)
#define swap(a, b) {int t; t = a; a = b; b = t; }

//Offsets into the persistent RAM of variables
#ifdef MX470
    #ifndef pin100
        #define RS_Pin_No 27
        #define WR_Pin_No 24
        #define RS_Pin 0x1000
        #define WR_Pin 0x0800
        #define clrport bclrport
        #define setport bsetport
    #else
        #define RS_Pin_No 18
        #define WR_Pin_No 19
        #define RS_Pin 0x100
        #define WR_Pin 0x200
        #define clrport eclrport
        #define setport esetport
    #endif
    #define port eport
    #define read eread
    #define trisinv etrisinv
    #define RSLo    {clrport=RS_Pin;}
    #define RSHi    {setport=RS_Pin;}
    #define WRLo    {clrport=WR_Pin;asm("NOP");}
    #define WRHi    {setport=WR_Pin;asm("NOP");}
#else
    #define RS_Pin_No 4
    #define WR_Pin_No 5
    #define RS_Pin 0x100
    #define WR_Pin 0x200
    #define clrport cclrport
    #define setport csetport
    #define port cport
    #define read cread
    #define trisinv ctrisinv
    #define RSLo    clrport=RS_Pin
    #define RSHi    setport=RS_Pin
    #define WRLo    clrport=WR_Pin
    #define WRHi    setport=WR_Pin
#endif
#define Both WR_Pin | RS_Pin
#define RDLo    (*(volatile unsigned int *)RDclrport)=RDpin
#define RDHi    (*(volatile unsigned int *)RDsetport)=RDpin
#define LANDSCAPE       1
#define PORTRAIT        2
#define RLANDSCAPE      3
#define RPORTRAIT       4

/*******************************************************************************
 *
 * Write Data to a register on the Chip
 *
 ******************************************************************************/
void write_command_data(unsigned int command, int data, ...){
   int i;
    #ifdef MX470 
        #ifndef pin100
            RSLo;
        #endif
    #endif
    #ifdef RGB565
        aclrport=0xFF;
    #endif
   va_list ap;
   va_start(ap, data);
   port=(command & 0x00FF) | WR_Pin;  WRLo;    WRHi; // RS low
    #ifdef MX470 
        #ifndef pin100
            RSHi;
        #endif
    #endif
   for(i = 0; i < data; i++) {
       port= (char)va_arg(ap, int) | Both;  WRLo;    WRHi; //RS high
   }
   va_end(ap);
}
/*******************************************************************************
 *
 * defines start/end coordinates for memory access from host to SSD1963
 * also maps the start and end points to suit the orientation
 *
 * This function is a modified version of the function inside the MMBasic Interpreter
 * for MM+ on 'MX470 chips
 *
*******************************************************************************/
void defineregion(long x, long y, long width,long height){ //SSD1963
    long x1=x,x2=x+width-1,y1=y,y2=y+height-1;
    unsigned long xstart,xend,ystart,yend,Vertical,Horizontal;
        #ifdef RGB565
            aclrport=0xFF;
        #endif
        #ifdef MX470 
            #ifndef pin100
                RSLo;
            #endif
        #endif
        if(HRes>VRes){
        Vertical=VRes;
        Horizontal=HRes;
    }
    else {
        Vertical=HRes;
        Horizontal=VRes;
    }
     if(Option->DISPLAY_ORIENTATION!=LANDSCAPE)goto isP;
              xstart = x1;
              xend = x2;
              ystart = y1;
              yend = y2;
              if(Option->LCD_CD>6){ //reverse for 7" displays
                xstart = (Horizontal - 1) - x2;
                xend = (Horizontal - 1) - x1;
              }
              goto setreg;
     isP:         
     if(Option->DISPLAY_ORIENTATION!=PORTRAIT)goto isRL;
              xstart = y1;
              xend = y2;
              ystart = (Vertical - 1) - x2;
              yend = (Vertical - 1) - x1;
              goto setreg;
    isRL:
    if(Option->DISPLAY_ORIENTATION!=RLANDSCAPE)goto isRP;
              xstart = (Horizontal - 1) - x2;
              xend = (Horizontal - 1) - x1;
              ystart = (Vertical - 1) - y2;
              yend = (Vertical - 1) - y1;
              if(Option->LCD_CD>6){//reverse for 7" displays
                xstart = x1;
                xend = x2;
              }
              goto setreg;
     isRP:
              xstart = (Horizontal - 1) - y2;
              xend = (Horizontal - 1) - y1;
              ystart = x1;
              yend = x2;
    setreg:
    port=0x22A ;WRLo;    WRHi; // RS low
    #ifdef MX470 
        #ifndef pin100
            RSHi;
        #endif
    #endif
    port=(xstart>>8 ) | Both;  WRLo;    WRHi; // RS HIGH
    port=(xstart) | Both;  WRLo;    WRHi; // RS HIGH
    port=(xend>>8 ) | Both;  WRLo;    WRHi; // RS HIGH
    port=(xend) | Both;  WRLo;    WRHi; // RS HIGH
    #ifdef MX470 
        #ifndef pin100
            RSLo;
        #endif
    #endif
    port=0x22B ;  WRLo;    WRHi; // RS low
    #ifdef MX470 
        #ifndef pin100
            RSHi;
        #endif
    #endif
    port=(ystart>>8 ) | Both;  WRLo;    WRHi; // RS HIGH
    port=(ystart) | Both;  WRLo;    WRHi; // RS HIGH
    port=(yend>>8 ) | Both;  WRLo;    WRHi; // RS HIGH
    port=(yend) | Both;  WRLo;    WRHi; // RS HIGH    RSHi;
}

int ReadRectangle(int x1, int y1, int width, int height, int pos, char *p ){
   int j,RDpin, RDsetport, RDclrport;
        RDpin=1<<GetPinBit(Option->LCD_CS);
        RDsetport=(unsigned int)GetPortAddr(Option->LCD_CS,LATSET);
        RDclrport=(unsigned int)GetPortAddr(Option->LCD_CS,LATCLR);

#ifdef RGB565
        j=width*height*2+pos;
#else
        j=width*height*3+pos;
#endif
        defineregion(x1,y1,width,height);
        #ifdef MX470 
            #ifndef pin100
                RSLo;
            #endif
        #endif
        port=0x22E ;  WRLo;    WRHi; // RS low
        #ifdef MX470 
            #ifndef pin100
                RSHi;
            #endif
        #endif
        trisinv=0xFF; //set pins to read
#ifdef RGB565
        atrisinv=0xFF;
#endif
        do { //read in the screen area to be overlayed
#ifndef RGB565
            RDLo;
            asm("NOP");asm("NOP");asm("NOP");
            RDHi;
            p[pos++]=(read & 0xFF);
            RDLo;
            asm("NOP");asm("NOP");asm("NOP");
            RDHi;
            p[pos++]=(read & 0xFF);
            RDLo;
            asm("NOP");asm("NOP");asm("NOP");
            RDHi;
            p[pos++]=(read & 0xFF);
#else
            RDLo;
            asm("NOP");asm("NOP");asm("NOP");
            RDHi;
            p[pos++]=(read & 0xFF);
            p[pos++]=(aread & 0xFF);
#endif
        } while(pos<j);
        trisinv=0xFF; //set pins to write
#ifdef RGB565
        atrisinv=0xFF;
#endif


}
int RestoreRectangle(int x1, int y1, int width, int height, int pos, char p[]){
    int i;
    defineregion(x1,y1,width,height);
    i=width*height;
    #ifdef MX470 
        #ifndef pin100
            RSLo;
        #endif
    #endif
    #ifdef RGB565
        aclrport=0xFF;
    #endif
    port=0x22C ;  WRLo;    WRHi; // RS low
    #ifdef MX470 
        #ifndef pin100
            RSHi;
        #endif
    #endif
#ifdef RGB565
    while (i--) {
        port=p[pos++] | Both;   asetport=p[pos]; aclrport=(p[pos++]^0xFF);
        WRLo;    WRHi;
    }
#else
    while(i--){
        port=p[pos++] | Both;    WRLo;    WRHi;
        port=p[pos++] | Both;    WRLo;    WRHi;
        port=p[pos++] | Both;    WRLo;    WRHi;
    }
#endif
    return pos;
}
/***********************************************************************************************
 * Display the bitmap of a char on the TFT panel
 *
 * This function is NEVER called by MMBasic programs
 *
 * This is optimised for the
 *      x, y - the top left of the char
 *      width, height - size of the char's bitmap
 *      scale - how much to scale the bitmap
 * 	fg, bg - foreground and background colour
 *      bitmap - pointer to the butmap
 *
 * This function is a modified version of the function inside the MMBasic Interpreter
 * for MM+ on 'MX470 chips
***********************************************************************************************/
void DrawBitmap_SSD1963(int x1, int y1, int width, int height, int scale, int fc, int bc, unsigned char *bitmap ){
    long i, j, k, m, n;
    unsigned int fhb,fmb, flb, bhb, bmb, blb, xtest, ytest;
    unsigned char *p;//,q[10];
    xtest=HRes;
    ytest=VRes;
    #ifdef RGB565
        aclrport=0xFF;
    #endif

    if(bc==1 && Option->LCD_CS!=0){ //special case of overlay text
    #ifdef RGB565
        p=GetMemory(width*scale*height*scale*2);
    #else
        p=GetMemory(width*scale*height*scale*3);
    #endif        
        ReadRectangle(x1,y1,width*scale,height*scale,0,p);
//        IntToStr(q,p[0],16);
//        MMPrintString(q);
//        putConsole(32);
//        IntToStr(q,p[1],16);
//        MMPrintString(q);
//        putConsole(32);
//        IntToStr(q,p[2],16);
//        MMPrintString(q);
//        putConsole(13);
    }
#ifndef RGB565        
    // convert the colours to 888 format
    fhb = (fc >> 16) | Both;
    fmb = (fc >> 8) | Both;
    flb = fc  | Both;
    bhb = (bc >> 16) | Both;
    bmb = (bc >> 8) | Both;
    blb = bc | Both;
#else
    fhb = ((fc >> 16) & 0b11111000) | ((fc >> 13) & 0b00000111);
    fmb = (((fc >>  5) & 0b11100000) | ((fc >>  3) & 0b00011111))|Both;
    flb=fhb^0xFF;
    bhb = ((bc >> 16) & 0b11111000) | ((bc >> 13) & 0b00000111);
    bmb = (((bc >>  5) & 0b11100000) | ((bc >>  3) & 0b00011111))|Both;
    blb=bhb^0xFF;

#endif
    n=0;
    defineregion(x1,y1,width*scale,height*scale);
    #ifdef MX470 
        #ifndef pin100
            RSLo;
        #endif
    #endif
    port=0x22C ;  WRLo;    WRHi; // RS low
    #ifdef MX470 
        #ifndef pin100
            RSHi;
        #endif
    #endif
    for(i = 0; i < height; i++) {                                   // step thru the font scan line by line
        for(j = 0; j < scale; j++) {                                // repeat lines to scale the font
            for(k = 0; k < width; k++) {                            // step through each bit in a scan line
                for(m = 0; m < scale; m++) {                        // repeat pixels to scale in the x axis
                    if(x1 + k * scale + m >= 0 && x1 + k * scale + m < xtest && y1 + i * scale + j >= 0 && y1 + i * scale + j < ytest) {  // if the coordinates are valid
                        if((bitmap[((i * width) + k)/8] >> (((height * width) - ((i * width) + k) - 1) %8)) & 1) {
#ifndef RGB565
                                port=fhb;    WRLo;    WRHi;
                                port=fmb;    WRLo;    WRHi;
                                port=flb;    WRLo;    WRHi;
                                n+=3;
#else
                                port=fmb;
                                asetport=fhb;aclrport=flb;
                                WRLo;    WRHi;
                                n+=2;
#endif
                        } else {
                            if(bc==1 && Option->LCD_CS!=0){
#ifndef RGB565
                                port=p[n++] | Both;    WRLo;    WRHi;
                                port=p[n++] | Both;    WRLo;    WRHi;
                                port=p[n++] | Both;    WRLo;    WRHi;
#else
                                port=p[n++] | Both;
                                asetport=p[n];aclrport=(p[n++]^0xFF);
                                WRLo;    WRHi;
#endif
                            } else {
#ifndef RGB565
                                port=bhb;    WRLo;    WRHi;
                                port=bmb;    WRLo;    WRHi;
                                port=blb;    WRLo;    WRHi;
                                n+=3;
#else
                                port=bmb;
                                asetport=bhb;aclrport=blb;
                                WRLo;    WRHi;
#endif
                            }
                        }
                    }
                }
            }
        }
    }
    if(bc==1 && Option->LCD_CS!=0) FreeMemory(p);
}

/*******************************************************************************
 *
 * Called by MMBasic Interpreter to draw a rectangle
 * on the SSD1963 panel
 *
 * This function is NEVER called by MMBasic programs
 *
 *
 ******************************************************************************/

void DrawRectangle_SSD1963(int x1, int y1, int x2, int y2, int fc){
    int t;
    long width, height;
    unsigned long i,ch,cm,cl;
    // make sure the coordinates are kept within the display area
    if(x2 <= x1) { t = x1; x1 = x2; x2 = t; }
    if(y2 <= y1) { t = y1; y1 = y2; y2 = t; }
    if(x1 < 0) x1 = 0; if(x1 >= HRes) x1 = HRes - 1;
    if(x2 < 0) x2 = 0; if(x2 >= HRes) x2 = HRes - 1;
    if(y1 < 0) y1 = 0; if(y1 >= VRes) y1 = VRes - 1;
    if(y2 < 0) y2 = 0; if(y2 >= VRes) y2 = VRes - 1;
    width=x2-x1+1;height=y2-y1+1;
    i=width*height;
    defineregion(x1,y1,width,height);
    #ifdef MX470 
        #ifndef pin100
            RSLo;
        #endif
    #endif
    #ifdef RGB565
        aclrport=0xFF;
    #endif
    port=0x22C ;  WRLo;    WRHi; // RS low
    #ifdef MX470 
        #ifndef pin100
            RSHi;
        #endif
    #endif
#ifdef RGB565
    ch = ((fc >> 16) & 0b11111000) | ((fc >> 13) & 0b00000111);
    cm = ((fc >>  5) & 0b11100000) | ((fc >>  3) & 0b00011111);
    port=cm | Both;   asetport=ch; aclrport=(ch^0xFF);
    while (i--) {
         WRLo;    WRHi;
    }
       
#else
   
    // convert the colours to 888 format
    ch = ((fc >> 16)) & 0xFF | Both; // make 8 bits to allow the optimisation to work
    cm = ((fc >> 8)) & 0xFF | Both;
    cl = fc & 0xFF | Both;
    if((ch==cl) && (ch==cm)){
       port=ch;    WRLo;    WRHi;
       WRLo;    WRHi;
       WRLo;    WRHi;
       i--;
       while (i--){
          WRLo;
          WRHi;
          WRLo;
          WRHi;
          WRLo;
          WRHi;
       } 
    }else {
        while (i--) {
            port=ch;    WRLo;    WRHi;
            port=cm;    WRLo;    WRHi;
            port=cl;    WRLo;    WRHi;
        }
    }
#endif
}
long triangles(int mode, int buffpos, char *buff, long  col, long x0, long y0, long x1, long y1,long x2, long y2) {//Cfunction
  long a, b, y, last;
  long  dx01,  dy01,  dx02,  dy02, dx12,  dy12,  sa, sb;
      
      if (y0>y1) {
        swap(y0, y1);
        swap(x0, x1);
      }
      if (y1>y2) {
        swap(y2, y1);
        swap(x2, x1);
      }
      if (y0>y1) {
        swap(y0, y1);
        swap(x0, x1);
      }
      if(y0==y2) { // Handle awkward all-on-same-line case as its own thing
        a=x0;
        b=x0;
        if(x1<a) {
          a=x1;
        } else {
          if(x1>b) b=x1;
        }
        if(x2<a) {
          a=x2;
        } else {
          if(x2>b)  b=x2;
        }
        if(mode==0){
            buffpos=ReadRectangle(a,y0,1,b-a+1,buffpos,buff);
        } else if(mode==1) {
            buffpos=RestoreRectangle(a,y0,1,b-a+1,buffpos,buff);
        } else {
            DrawRectangle_SSD1963(a,y0,a,b-a+y0,col);
        }
      } else {
        dx01=x1-x0;  dy01=y1-y0;  dx02=x2-x0;  dy02=y2-y0; dx12=x2-x1;  dy12=y2-y1;  sa=0; sb=0;
        if(y1==y2) {
            last=y1; //Include y1 scanline
        } else {
            last=y1-1; // Skip it
        }
        for (y=y0;y<=last;y++){
            a=x0+sa / dy01;
            b=x0+sb / dy02;
            sa=sa+dx01;
            sb=sb+dx02;
            a=x0+(x1-x0) * (y-y0) / (y1-y0);
            b=x0+(x2-x0) * (y-y0) / (y2-y0);
            if(a>b)swap(a,b);
            if(mode==0){
                buffpos=ReadRectangle(a,y,b-a+1,1,buffpos,buff);
            } else if(mode==1) {
                buffpos=RestoreRectangle(a,y,b-a+1,1,buffpos,buff);
            } else {
                DrawRectangle_SSD1963(a,y,b,y,col);
            }
        }
        sa=dx12 * (y-y1);
        sb=dx02 * (y-y0);
        while (y<=y2){
            a=x1+sa / dy12;
            b=x0+sb / dy02;
            sa=sa+dx12;
            sb=sb+dx02;
            a=x1+(x2-x1) * (y-y1) / (y2-y1);
            b=x0+(x2-x0) * (y-y0) / (y2-y0);
            if(a>b) swap(a,b);
            if(mode==0){
                buffpos=ReadRectangle(a,y,b-a+1,1,buffpos,buff);
            } else if(mode==1) {
                buffpos=RestoreRectangle(a,y,b-a+1,1,buffpos,buff);
            } else {
                DrawRectangle_SSD1963(a,y,b,y,col);
            }
            y=y+1;
            }
        }

  return buffpos;
}

__attribute__((noinline)) void getFPC(void *a, void *b, volatile unsigned int *c) 
     { 
         *c = (unsigned int) (__builtin_return_address (0) - (b -a)) ;      
     } 
void pstring(const char *s){
    volatile unsigned int libAddr ; 
    getFPC(NULL,&&getFPCLab,&libAddr) ; // warning can be ignored, stupid editor 
    getFPCLab: { } 
    unsigned char  * testData    = (unsigned char *)((void *)s + libAddr );
    MMPrintString(testData);
}

/*******************************************************************************
 *
 * SSD1963_V44 : Initialise the CFunction Driver Sub-System
 *
 * Function called to initialise the driver SubSystem
 *
 * SSD1963_V44 is ALWAYS called from an MMBasic program
 * On exit, vectors DrawRectangleVector, and DrawBitmapVector will
 * be set to point to the CFunctions DrawRectangle_SSD1963 and
 * DrawBitmap_SSD1963 respectively
 *
 * Input Arguments
 *    MyAddress:    The Address of the CFunction, use PEEK(CFUNADDR SSD1963_V44)
 *    RST:  The pin connected to the Reset input on the display
 *    orientation:
 *       is 0 for Portrait orientation, 1 for reverse portrait, 2 for landscape, 3 for reverse landscape
 *    size: The display size 4, 5, or 7"
 *
 * Exit Value   Version
 *
 * Note, For added performance, it may be desirable to create a
 * separate WriteData function which doen't asser/de-asser CS for each byte and
 * instead use block CSLo/CSHi around data transfers. These "block" CSLo/CSHi
 * statements have been left in the code as comments.
 * 
 ******************************************************************************/
//CFunction Driver_SSD1963
long long main(char *orientation, unsigned char *p1, long long p2[], long long p3[], long long p4[], long long p5[], long long p6[],long long p7[], long long p8[]){
    int t,HorizontalRes,VerticalRes;
    unsigned int p[16];
    int j;
    if(Option->LCD_CS){
        PinSetBit(Option->LCD_CS, LATSET);
    }
    volatile unsigned int libAddr ; 
    getFPC(NULL,&&getFPCLab,&libAddr) ; // warning can be ignored, stupid editor 
    getFPCLab: { } 
    
    if(*orientation>=0 && *orientation<=4){
    Option->DISPLAY_ORIENTATION=*orientation;
    Option->LCD_CS=0;
#ifndef MX470
    if(p2[0])Option->LCD_CS=p2[0];
#else
    if(*p1)Option->LCD_CS=*p1;
#endif
    Option->DISPLAY_ORIENTATION=*orientation;
    Option->LCD_CD=*p2; //save the display size as the display type for use in the driver routines

#ifdef MX470
    #ifndef pin100
        Option->LCD_Reset=28;
        p[0]=60;
        p[1]=61;
        p[2]=62;
        p[3]=63;
        p[4]=64;
        p[5]=1;
        p[6]=2;
        p[7]=3;
    #else
        Option->LCD_Reset=42;
        p[0]=93;
        p[1]=94;
        p[2]=98;
        p[3]=99;
        p[4]=100;
        p[5]=3;
        p[6]=4;
        p[7]=5;
        p[8]=17;
        p[9]=38;
        p[10]=58;
        p[11]=59;
        p[12]=60;
        p[13]=61;
        p[14]=91;
        p[15]=92;
    #endif

#else    
    Option->LCD_Reset=*p1;
    p[0]=25;
    p[1]=26;
    p[2]=27;
    p[3]=36;
    p[4]=37;
    p[5]=38;
    p[6]=2;
    p[7]=3;
#endif    
    for(j=0;j<8;j++){
        ExtCfg(p[j],EXT_DIG_OUT,0);ExtCfg(p[j],EXT_BOOT_RESERVED,0);
        PinSetBit(p[j],LATCLR);
    }
#ifdef RGB565
    for(j=8;j<16;j++){
        ExtCfg(p[j],EXT_DIG_OUT,0);ExtCfg(p[j],EXT_BOOT_RESERVED,0);
        PinSetBit(p[j],LATCLR);
    }
#endif
    //Control Signals
    ExtCfg(RS_Pin_No,EXT_DIG_OUT,0);ExtCfg(RS_Pin_No,EXT_BOOT_RESERVED,0);
    RSHi;

    ExtCfg(WR_Pin_No,EXT_DIG_OUT,0);ExtCfg(WR_Pin_No,EXT_BOOT_RESERVED,0);
    WRHi;

    ExtCfg(Option->LCD_Reset,EXT_DIG_OUT,0);ExtCfg(Option->LCD_Reset,EXT_BOOT_RESERVED,0);
    PinSetBit(Option->LCD_Reset, LATSET);

    if(Option->LCD_CS){
        ExtCfg(Option->LCD_CS, EXT_DIG_OUT,0);
        ExtCfg(Option->LCD_CS, EXT_BOOT_RESERVED,0);
        PinSetBit(Option->LCD_CS, LATSET);
    }

    //Save DisplayMetrics
    if(Option->LCD_CD<5){
       HorizontalRes=480;
//       if(Option->DISPLAY_ORIENTATION==1) VerticalRes=272*3; //3 frame buffers
       /*else*/ VerticalRes=272;
    } else {
      HorizontalRes=800;
      VerticalRes=480;
    }

    //Reset the SSD1963
    PinSetBit(Option->LCD_Reset,LATSET);
    uSec(10000);
    PinSetBit(Option->LCD_Reset,LATCLR);
    uSec(10000);
    PinSetBit(Option->LCD_Reset,LATSET);
    uSec(10000);

    if(Option->LCD_CD<5) {
        write_command_data(0xE2,3,0x23,0x02,0x54);
        write_command_data(0x33,6,0,0,1,16,0,0); //set scroll area
    } else {
	if(Option->LCD_CD<6){
             write_command_data(0xE2,3,0x1E,0x02,0x54);
        } else {
             write_command_data(0xE2,3,0x23,0x02,0x04);
        }
    }

    write_command_data(0xE0,1,0x01); uSec( 10000); 		// PLL enable
    write_command_data(0xE0,1,0x03);	uSec( 10000);
    write_command_data(0x01,0); uSec( 100000);		// software reset

    if(Option->LCD_CD<5) {
	write_command_data(0xE6,3,0x01,0x1F,0xFF);
	write_command_data(0xB0,7,0x20,0x00,0x01,0xDF,0x01,0x0F,0x00);		//LCD SPECIFICATION
	write_command_data(0xB4,8,0x02,0x13,0x00,0x08,0x2B,0x00,0x02,0x00);
	write_command_data(0xB6,7,0x01,0x20,0x00,0x04,0x0C,0x00,0x02);
    } else {
        if (Option->LCD_CD<6){
            write_command_data(0xE6,3,0x03,0xFF,0xFF);
            write_command_data(0xB0,7,0x24,0x00,0x03,0x1F,0x01,0xDF,0x00);		//LCD SPECIFICATION
        } else {
            write_command_data(0xE6,3,0x04,0x93,0xE0);
            write_command_data(0xB0,7,0x00,0x00,0x03,0x1F,0x01,0xDF,0x00);		//LCD SPECIFICATION
        }
        write_command_data(0xB4,8,0x03,0xA0,0x00,0x2E,0x30,0x00,0x0F,0x00);
        write_command_data(0xB6,7,0x02,0x0D,0x00,0x10,0x10,0x00,0x08);
    }

    write_command_data(0xBA,1,0x0F);
    write_command_data(0xB8,2,0x07,0x01);
    write_command_data(0x36,1,0x00);		//rotation
#ifdef RGB565
    write_command_data(0xF0,1,0x03);		//PIXEL data interface - 16-bit RGB565
    write_command_data(0x3A,1,0b01010000);		//PIXEL data interface RGB565
#else
    write_command_data(0xF0,1,0x00);		//PIXEL data interface - 8-bit
    write_command_data(0x3A,1,0b01110000);		//PIXEL data interface 24-bit
#endif
    uSec( 1000);	
    write_command_data(0x26,1,0x01);  uSec( 10000);		//gamma curve
    write_command_data(0x29,0);		//display on
    write_command_data(0xBE,6,0x06,0xF0,0x01,0xF0,0x00,0x00);		//set PWM for B/L
    write_command_data(0xd0,1,0x0D);
    //Set Hres and VRes
    if(Option->DISPLAY_ORIENTATION==PORTRAIT){
        if (Option->LCD_CD<6) t=0xA0;
        else t=0xE0;
    }
    if(Option->DISPLAY_ORIENTATION==RPORTRAIT){
        if (Option->LCD_CD<6) t=0x60;
        else t=0x20;
    }
    if(Option->DISPLAY_ORIENTATION==LANDSCAPE){
        if (Option->LCD_CD<6) t=0x00;
        else t=0x40;
    }
    if(Option->DISPLAY_ORIENTATION==RLANDSCAPE){
        if (Option->LCD_CD<6) t=0xC0;
        else t=0x80;
    }
    write_command_data(0x36,1,t);
    if(Option->DISPLAY_ORIENTATION&1){
        HRes=HorizontalRes;
        VRes=VerticalRes;
    } else {
        VRes=HorizontalRes;
        HRes=VerticalRes;
    }

    //Save the current values of the Vectors

    //Set the DrawRectangle vector to point to our function
    DrawRectangleVector= (unsigned int)&DrawRectangle_SSD1963 + libAddr;

    //Set the DrawBitmap vector to point to our function
    DrawBitmapVector=(unsigned int)&DrawBitmap_SSD1963 + libAddr;

    //CLS
    if(Option->LCD_CD<5) {
        DrawRectangle(0,0,HRes-1,VRes*3-1,0x000000);
    } else {
        DrawRectangle(0,0,HRes-1,VRes-1,0x000000);
    }
#ifdef MX470
#ifndef pin100
      static const char startup[]="SSD1963_64 driver loaded\r\n";
#else
      static const char startup[]="SSD1963_100 driver loaded\r\n";
#endif
#else    
       static const char startup[]="SSD1963_44 driver loaded\r\n";
#endif
    pstring(startup);
    } else if(*orientation==5){
        int num=p1[0],a=1;
        unsigned char ch,cm;
        if(p2!=NULL){
            num=*p2-1;
            a=0;
        }
        #ifdef RGB565
        while (a<num) {
//            port=p1[a++] | Both;   asetport=p1[a]; aclrport=(p1[a++]^0xFF);
            ch=p1[a++] & 0b11111000;
            ch|=p1[a]>>5;
            cm=(p1[a++]<<3) & 0b11100000;
            cm|=(p1[a++]>>3);
            port=cm | Both;   asetport=ch; aclrport=(ch^0xFF);
            WRLo;    WRHi;
        }
        #else
        while(a<num){
            port=p1[a++] | Both;    WRLo;    WRHi;
            port=p1[a++] | Both;    WRLo;    WRHi;
            port=p1[a++] | Both;    WRLo;    WRHi;
        }
        #endif
    } else if(*orientation==6){
            return ReadRectangle(*p2, *p3, *p4, *p5, *p6, p1 );
    } else if(*orientation==7){
            return RestoreRectangle(p2[0], p3[0], p4[0], p5[0], p6[0], p1 );
    } else if(*orientation==8){
        defineregion(*p1, *p2, *p3, *p4);
        #ifdef MX470 
            #ifndef pin100
                RSLo;
            #endif
        #endif
        #ifdef RGB565
            aclrport=0xFF;
        #endif
        port=0x22C ;  WRLo;    WRHi; // RS low
        #ifdef MX470 
            #ifndef pin100
                RSHi;
            #endif
        #endif
        return 8;
    } else if(*orientation==9){
       if(Option->DISPLAY_ORIENTATION==1){
           VRes=272*3;
           j=(*p1-1) * 272; 
           write_command_data(0x37,2,j>>8,j&0xFF); 
       }
    } 
#ifdef RGB565
    else if(*orientation==10){
        int num=p1[0],a=1,astore;
        astore=aport & 0xFF00;
        if(p2!=NULL){
            num=*p2-1;
            a=0;
        }
        while (a<num) {
            aport=p1[a++]|astore;
            port=p1[a++] | Both;   
            WRLo;    WRHi;
        }
#endif
    } else { //must be triangles
        int count=*orientation & 0x7F;
        int i=0,j,buffpos=0;
        while(i<count && p2[i]<0){ //repeat for any restores
            buffpos=triangles(1,buffpos,p1,p2[i],p3[i],p4[i],p5[i],p6[i],p7[i],p8[i]);
            i++;
        }  
        buffpos=0;
        j=i;
        while(i<count && p2[i]>=0){ //repeat for any store and write commands
            buffpos=triangles(0,buffpos,p1,p2[i],p3[i],p4[i],p5[i],p6[i],p7[i],p8[i]);
            i++;
        }
        i=j;
        while(i<count && p2[i]>=0){ //repeat for any store and write commands
            buffpos=triangles(2,buffpos,p1,p2[i],p3[i],p4[i],p5[i],p6[i],p7[i],p8[i]);
            i++;
        }
       return buffpos>>3;        

    }
}
